/* ========================================================================= */
/* =                        FONCTIONS PRINCIPALES                          = */
/* ========================================================================= */

/* *** PROTOTYPES *** */
#include "altaz.h"

/* ============================================== */
/* = Fonction   : conversion degres -> radians  = */
/* ============================================== */
double deg2rad( double alpha ) { return( (alpha*PI) / 180.0 ); }

/* ============================================== */
/* = Fonction   : conversion radians -> degres  = */
/* ============================================== */
double rad2deg( double alpha ) { return( (alpha*180) / PI ); }

/* ====================================================== */
/* = Fonction    : Extrait la partie entier d'un float = */
/* = Entree      : double                               = */
/* = Sortie      : double                               = */
/* ====================================================== */
double Int_Float(double number)
{
 double v = 0.0;
 modf(number, &v);
 return(v);
}

/* ====================================================== */
/* = Fonction    : Day -> Decimal                       = */
/* = Entree      : D,H,M,S                              = */
/* = Sortie      : double decimal date DD.dd            = */
/* ====================================================== */
double Day2Float(float Day, float Hour, float Min, float Sec)
{
 double DD = 0.0;
 DD  = Day;
 DD += Hour/24.0;
 DD += Min/1440.0;
 DD += Sec/86400.0;
 return(DD); 
}

/* ====================================================== */
/* = Fonction    : convert DMS -> float                 = */
/* = Entree      : Deg, Min, Sec values                 = */
/* = Sortie      : double                               = */
/* ====================================================== */
double DMS2Float(int deg, unsigned int min, unsigned int sec)
{
 double d = 0.0;
 d+=(double)fabs(deg);
 d+=(double)min/60.0;
 d+=(double)sec/3600.0;
 (deg < 0.0) ? (d = d*(-1.0)) : d;
 return(d); 	
}

/* ====================================================== */
/* = Fonction    : DecimalDegrees to DMS                = */
/* = Entree      : float degrees                        = */
/* = Sortie      : float struct D M S pointed by input  = */
/* ====================================================== */
void Float2DMS(double dec, struct DMS *dms)
{
 dms->d = Int_Float(dec); // 1 hour equ 15 degrees
 dms->m = Int_Float(fabs(dec-dms->d)*60.0);
 dms->s = ((fmod(fabs(dec-dms->d),60.0)*60.0)-dms->m)*60.0;
 if ( dms->s > 59.0 )
 {
 	dms->s = 0.0;
 	dms->m++;
 }
}

/* ============================================== */
/* = Fonction    : Temps de pose maximum        = */
/* = Entree      : focale;taille pixel, hauteur = */
/* = Sortie      : Temps                        = */
/* ============================================== */
inline float b_max( float f, float p, float h)
{
 return( ((1.0/cos(deg2rad(h)))*((180.0*3600.0)/(ROT_HORAIRE*PI))*(p/f))/1000.0 );
}

/* ============================================== */
/* = Fonction    : affiche aide                 = */
/* ============================================== */
void aff_hlp()
{
 char help[] = "Evaluation de la rotation de champ d'une monture Altazimutale\nGeneration d'un fichier ROT.LST utilisable par Iris (astrosurf.com/buil)\n\taltaz [-param] valeur [/option] valeur(s)\nObligatoires :\n\t-lat : latitude du lieu d'observation\n\t-lon : longitude du lieu d'observation (+:Ouest/-:Est de Greenwich)\n\t-a   : azimuth objet\n\t-h   : hauteur objet\n\t-i   : nombre images par seconde\n\t-n   : nombre total d'images\n\t-x   : abscisse du centre image\n\t-y   : ordonnee du centre image\nOptionnels :\n\t/P   : repertoire de travail d'Iris (ou courant par defaut)\n\t/T   : rotation dans le sens trigonometrique\n\t/S   : affiche le resultat a l'ecran sans creer de fichier\n\t/O   : ajoute un angle au calcul (degres decimal, sens horaire)\n\t/FP  : Focale ET Taille pixel(mm), calcul le temps de pose Max.\n\t/E   : Num_Planete : 3,venus | 4,mars | 5,jupiter | 6,saturne\n\t+ /JJ.jj : Jour + /MM : Mois + /AA : Annee   (Temps Universel)\n\t+ /HH : Heures + /MIN : Minutes + /SEC.ss : Secondes\n\nExemple: altaz -lat 49.5 -a 145.21 -h 61.7 -i 5 -n 350 -x 128 -y 128 /P c:\\wks\n\tVersion ";
 char sgn[] = " (logiciel libre), http://yann.duchemin.free.fr/";
 printf("%s%s%s",help,VERSION,sgn);
}

/* ============================================== */
/* = Fonction    : affectation argumentation    = */
/* = Entree      : rang arg, pointeur arg       = */
/* = Sortie      : code err                     = */
/* ============================================== */
int init_str(int swh, char *dta, ROT_STRUCT *p_rot)
{
 float cnf = atof(dta);
 char *dtb = dta;
 char tab[255];

 strcpy(tab, dta);
 strcat(tab,"\n");

 switch( swh )
 {
  case LATITUDE :
   p_rot->lat = cnf;
   break;

  case LONGITUDE :
   p_rot->lon = cnf;
   break;

  case AZIMUTH :
  p_rot->az = cnf;
   break;

  case HAUTEUR :
  p_rot->ht = cnf;
   break;

  case IMG_SEC :
  p_rot->img = cnf;
   break;

  case IMG_TOT :
  p_rot->nimg = cnf;
   break;

  case CENTER_X :
  p_rot->cx = cnf;
   break;

  case CENTER_Y :
  p_rot->cy = cnf;
   break;

  case DIRECTORY :
   strcpy(p_rot->dir, dta);
   break;

  case TRIGO :
  p_rot->trig = 1;
   break;

  case ECRAN :
  p_rot->scr = 1;
   break;

  case OFFSET :
  p_rot->offs = cnf;
   break;

  case FOCALE :
  p_rot->f = cnf;
   dtb+=strlen(tab);
  p_rot->p = atof(dtb);
   break;

  case OBJET :
  p_rot->obj = atoi(dta);
   break;

  case JOUR :
  p_rot->jj = (double)cnf;
   break;

  case MOIS :
  p_rot->mm = atoi(dta);
   break;

  case ANNEE :
  p_rot->aa = atoi(dta);
   break;

  case HEURE :
  p_rot->hrs = atoi(dta);
   break;

  case MINUTE :
  p_rot->min = atoi(dta);
   break;

  case SECONDE :
  p_rot->sec = cnf;
   break;

  case HELP :
  aff_hlp();
  exit(-1);
  break;

  default :
   return(1);

 }
 return(0);
}

/* ============================================== */
/* = Fonction    : Traitement des donnees       = */
/* ============================================== */
void trt_dta(struct ROT_STRUCT *p_rot)
{
 double Rch, Ri, Rit = 0.0;
 double Bmax = 0.0, B = 0.0;
 double dureeT = 0.0;
 FILE *dest;
 double i = 0.0, j = 0.0;

 char fname[255], buf[512];
 char *p_buf = buf;

 struct DMS f2dms;
 struct DMS *p_f2dms = &f2dms;

 POSITION p;

 double t=0.0;

 strcpy(fname, strcat(p_rot->dir,".\\rot.lst"));

 if( (p_rot->hrs != -1) || (p_rot->min != -1) || (p_rot->sec != -1.0) )
 {
  p_rot->jj += Day2Float(0.0, p_rot->hrs, p_rot->min, p_rot->sec);
 }
 else
 {
  j = p_rot->jj;
  j = Int_Float(j);
  Float2DMS(p_rot->jj, p_f2dms);
  p_rot->hrs = p_f2dms->d;
  p_rot->min = p_f2dms->m;
  p_rot->sec = p_f2dms->s;
 }

 if(p_rot->obj != -1)
 {
  if( p_rot->az != -1.0) { puts("-a integre par le calcul!"); p_rot->az = -1.0; }
  if( p_rot->ht != -1.0) { puts("-h integre par le calcul!"); p_rot->ht = -1.0; }

  p = CalculPosition(p_rot);

  printf("\nDate debut integration      : %02.6f / %02d / %d (%02d:%02d:%02.2f)\n",p_rot->jj,p_rot->mm,p_rot->aa,p_rot->hrs,p_rot->min,p_rot->sec);
  printf("Date Julienne               : %02.8f\n",JJ(p_rot->jj, p_rot->mm, p_rot->aa, p_rot->hrs, p_rot->min, p_rot->sec));
  printf("Estimation objet            : (%d) ",p_rot->obj);
  switch( p_rot->obj )
  {
   case MERCURE : printf("MERCURE");	break;
   case VENUS   : printf("VENUS");	break;
   case MARS    : printf("MARS");	break;
   case SATURNE : printf("SATURNE");	break;
   case JUPITER : printf("JUPITER");	break;
   default 	: printf("UTILISATEUR");
  }
  printf("\n");

  printf("Latitude du lieu            : %.02f ( ",p_rot->lat);
  Float2DMS(p_rot->lat, p_f2dms);
  printf("%.0fd %.0fm %.02fs )\n",p_f2dms->d, p_f2dms->m, p_f2dms->s);

  printf("Longitude du lieu           : %.02f ( ",p_rot->lon);
  Float2DMS(p_rot->lon, p_f2dms);
  printf("%.0fd %.0fm %.02fs )\n",p_f2dms->d, p_f2dms->m, p_f2dms->s);

  printf("Ascension Droite            : %.04f ( ",p.Asc);
  Float2DMS(p.Asc, p_f2dms);
  printf("%.0fh %.0fm %.02fs )\n",p_f2dms->d, p_f2dms->m, p_f2dms->s);

  printf("Declinaison objet           : %.04f ( ",p.Dec);
  Float2DMS(p.Dec, p_f2dms);
  printf("%.0fd %.0fm %.02fs )\n",p_f2dms->d, p_f2dms->m, p_f2dms->s);

  printf("Azimuth                     : %.04f ( ",p_rot->az);
  Float2DMS(p_rot->az, p_f2dms);
  printf("%.0fd %.0fm %.02fs )\n",p_f2dms->d, p_f2dms->m, p_f2dms->s);

  printf("Hauteur                     : %.04f ( ",p_rot->ht);
  Float2DMS(p_rot->ht, p_f2dms);
  printf("%.0fd %.0fm %.02fs )\n",p_f2dms->d, p_f2dms->m, p_f2dms->s);
 }

 dureeT = p_rot->nimg/p_rot->img;

 if( p_rot->f && (p_rot->p != -1) )
 {
  Bmax = b_max(p_rot->f, p_rot->p, p_rot->ht); //tps pose max en secondes
  printf("Focale                      : %.2f mm\n",p_rot->f);
  printf("Taille pixel                : %.2f microns\n",p_rot->p);
  printf("-> La pose par image est de : %.3f sec.\n",dureeT/p_rot->nimg);
  printf("-> La pose maximum est de   : %.3f sec. par image\n",Bmax);
  printf("-> Echantillonnage          : %.3f arcsec. par pixel\n",((((180.0/PI)*3600.0)/1000.0)*p_rot->p)/p_rot->f );
 }

 p_rot->sec = p_rot->sec - 1.0; // pour incrementation dans RImage

 // calcul de l'angle pour chaque image
 Ri = RImage(p_rot);
 Rch = rot_moy(p_rot->lat, p_rot->az, p_rot->ht);  // premiere estimation de la rotation de champ
 printf("-> L\'angle horaire est de   : %.3f Deg (%.3f Deg par sec.)\n", Rch, Rch/3600.0);
 printf("-> L\'angle moyen est de     : %.3f Deg par image\n",Ri);
 printf("-> L\'angle evalue est de    : %.3f Deg pour %.2f sec.\n", Ri*p_rot->nimg, dureeT);

 if( !(p_rot->offs == 0) )
 {
  printf("-> Un Angle de %f Deg. est ajoute a la sequence\n", p_rot->offs);
  if( !(p_rot->trig) ) Rit += (fabs(p_rot->offs)*(-1));
  else Rit += fabs(p_rot->offs);
 }

 j = 0.0; // on reinitialise cette varible precedement utilisee, pour compter le nombre total d'images
 B = (p_rot->nimg/p_rot->img)/p_rot->nimg;  // temps de pose theorique par image

 if( !(p_rot->scr) && ( (dest = fopen(fname, "wt"))!=NULL ) )
 {
  while( j < p_rot->nimg )
  {
   for(i=0.0; i<B ; i+=B, Rit+=Ri)
   {
    sprintf(buf, "%.3f", p_rot->cx);
    fwrite(buf, strlen(p_buf), 1, dest);
    fwrite(" ", 1, 1, dest);
    sprintf(buf, "%.3f", p_rot->cy);
    fwrite(buf, strlen(p_buf), 1, dest);
    fwrite(" ", 1, 1, dest);
    sprintf(buf, "%f", Rit);
    fwrite(buf, strlen(p_buf), 1, dest);
    fwrite("\n", 1, 1, dest);
   }
   // calcul de l'angle pour chaque image
   Ri = RImage(p_rot);
   j++;
  }
  fclose(dest);
  puts("-> Fichier rot.lst genere, utilisez la commande Iris : file_rot\n");
 }
 else
 {
  if(p_rot->obj != -1) { puts("\n Jour      Mois  Annee(Hr:Mn:Sec ) Angle: DMS D.d    Alpha: DMS D.d        Delta: DMS D.d      Azimuth: DMS D.d     Hauteur: DMS D.d"); }
  do
  {
   for(i=0.0; i<B; i+=B, Rit+=Ri)
   {
    if(p_rot->obj != -1)
    {
     printf("%.6f / %02d / %d (%02d:%02d:%02.2f) %.3f", p_rot->jj, p_rot->mm, p_rot->aa, p_rot->hrs, p_rot->min, p_rot->sec, Rit);
     Float2DMS(p_rot->asc, p_f2dms);
     printf(" %.0f:%.0f:%.02f ",p_f2dms->d, p_f2dms->m, p_f2dms->s);
	 printf(" %.6f ",p_rot->asc);
     Float2DMS(p_rot->dec, p_f2dms);
     printf(" %.0f:%.0f:%.02f ",p_f2dms->d, p_f2dms->m, p_f2dms->s);
	 printf(" %.6f ",p_rot->dec);
     Float2DMS(p_rot->az, p_f2dms);
     printf("%.0f:%.0f:%.02f ",p_f2dms->d, p_f2dms->m, p_f2dms->s);
	 printf(" %.6f ",p_rot->az);
     Float2DMS(p_rot->ht, p_f2dms);
     printf("%.0f:%.0f:%.02f ",p_f2dms->d, p_f2dms->m, p_f2dms->s);
	 printf(" %.6f\n",p_rot->ht);
    }
    else
    {
     printf("%.03f %.03f %.03f\n",p_rot->cx, p_rot->cy, Ri);     
    }
   }
   // calcul de l'angle pour chaque image
   Ri = RImage(p_rot);
   j++;
//----------------------------------------------------------------------- TEST
// printf(">%f %f %02d:%02d:%02.2f\n",Rit,j,p_rot->hrs,p_rot->min,p_rot->sec);
//----------------------------------------------------------------------- FIN TEST
  }while( j < p_rot->nimg );
 }

}

/* ============================================== */
/* = Fonction    : initialisation de structure  = */
/* ============================================== */
void raz_dta(struct ROT_STRUCT *p_rot)
{
 p_rot->lat = 0.0;
 p_rot->lon = 0.0;
 p_rot->az  = -1.0;
 p_rot->ht  = -1.0;
 p_rot->cx  = 0.0;
 p_rot->cy  = 0.0;
 p_rot->offs= 0.0;
 p_rot->img = 0.0;
 p_rot->nimg= 0.0;
 p_rot->f   = 0.0;
 p_rot->p   = 0.0;
 p_rot->b   = 0.0;
 p_rot->trig= 0;
 p_rot->scr = 0;
 p_rot->obj = -1;
 p_rot->jj  = 1.5; 
 p_rot->mm  = 1;
 p_rot->aa  = 2000;
 p_rot->hrs = -1;
 p_rot->min = -1;
 p_rot->sec = -1.0;
 strcpy(p_rot->dir,"");
}

/* ============================================== */
/* = Fonction    : principale                   = */
/* ============================================== */
int main(int argc, char *argv[], char *envp[])
{

 int i = 0, j = 0;

 // structures de donnees
 struct ROT_STRUCT rotation;
 struct ROT_STRUCT *p_rot = &rotation;

 // definition des options de ligne de commandes
 static char *opt[] = {
		    "-lat",	// 0 latitude
		    "-lon",	// 1 longitude
		    "-a",	// 2 azimuth
		    "-h",	// 3 hauteur
		    "-i",	// 4 nbr img par sec
		    "-n",	// 5 nbr total img
		    "-x",	// 6 centre img X
		    "-y",	// 7 centre img Y
		    "/P",	// 8 path
		    "/T",	// 9 rotation sens trigo
		    "/S",	//10 Affichage ecran uniquement
		    "/O",	//11 Offset a ajouter au calcul
		    "/FP",	//12,122 focale et taille d'un pixel en microns
		    "/E",	//13 calcule l'ephemeride planetaire et l'applique a toutes les images
		    "/JJ",	//14 jour
		    "/MM",	//15 mois
		    "/AA",	//16 annee
		    "/HRS",	//17 heure
		    "/MIN",	//18 minute
		    "/SEC",	//19 seconde
		    "/?"	//20 aide
	          };

 // initialisation des donnees
 raz_dta(p_rot);

 if(argc > 1)	// 1 = nom programme sans options
 {
  for( i=1; i<argc; i++)
  {
   for( j=0; j<NB_ARG; j++)
   {
    if( !(strcmp(argv[i], opt[j])) )
    {
     if(argv[i+1] != NULL)
     {
      if( init_str(j, argv[i+1], p_rot) ) { exit(-1); }
     }
     else 
     {
      if( init_str(j, "", p_rot) ) { exit(-1); }
     }
    }
   }
  }
  trt_dta(p_rot);
 }
 else
 {
//  for( int s=0; envp[s] != NULL; s++)
//  {
//   printf("%s >%d[%s]\n",argv[0],s,envp[s]);
//  }
  aff_hlp();	// parametres absents
  return(-1);
 }
 return(0);
}
