/* ========================================================================= */
/* =                   FONCTIONS DE CALCULS DE POSITIONS                   = */
/* ========================================================================= */

#include "altaz.h"

/* ====================================================== */
/* = Function    : Year Validation                      = */
/* = Input       : int year                             = */
/* = Output      : 1 if leap, if not 0                  = */
/* ====================================================== */
int VerifyYearB(int year)
{
 return (((year % 4 == 0) && ( (!(year % 100 == 0)) || (year % 400 == 0))) ? 1 : 0 );
}

/* ====================================================== */
/* = Function    : Nb Days For a Month                  = */
/* = Input       : int Month, int year                  = */
/* = Output      : number of day                        = */
/* ====================================================== */
int NbDaysFMonth(int month, int year)
{
 int DMM[] = {31,28,31,30,31,30,31,31,30,31,30,31};
 // month = month - 1 because tabs start at zero ;o)
 if( (VerifyYearB(year)) )
  DMM[1] = 29;		// feb have 29 days
 return(DMM[month-1]);
}

/* ============================================== */
/* = Fonction    : angle de rotation objet      = */
/* = Entree      : latitude, azimuth, hauteur   = */
/* = Sortie      : angle                        = */
/* ============================================== */
float rot_moy( float lat, float azo, float deco)
{
 return( ( (ROT_HORAIRE * cos(deg2rad(lat))) * cos(deg2rad(azo)) / cos(deg2rad(deco)) ) );
}

/* ====================================================== */
/* = Fonction    : Calcul du jour Julien                = */
/* = Entree      : jour, mois, annee, hrs, min, sec     = */
/* = Sortie      : float                                = */
/* ====================================================== */
double JJ(double jour, int mois, int annee , int hrs, int min, float sec)
{
 double A = 0.0,	// A & B are correction for the Julian's calendar reform
        B = 0.0;	// In other terms we're always in the gredorian's calend
 int Y = annee,
     M = mois,
     D = jour;
 if( M < 3)
 {
  M +=12;
  Y -= 1;
 }
 A = Int_Float( (float)Y / 100.0 );
 B = 2 - A + Int_Float( A / 4.0);
 B += Int_Float( 365.25 * (float)Y );				// C term
 B += Int_Float( 30.6001 * ( (float)M + 1.0) );			// D term
 B += (Day2Float( Int_Float(D), hrs, min, sec) + 1720994.5);	// A+B+C+D terms
 return(B);
}

/* ====================================================== */
/* = Function    : Sideral Time at Greenwich in hours   = */
/* = Input       : float JD                             = */
/* = Output      : float SDT                            = */
/* ====================================================== */
double STG_H(struct ROT_STRUCT *p)
{
 double S = 0.0, T = 0.0, T0 = 0.0, JD = 0.0, UT = 0.0;

 JD = JJ(p->jj, p->mm, p->aa, 0, 0, 0.0);
 UT = DMS2Float(p->hrs, p->min, p->sec);
 S = JD - 2451545.0;
 T = S/36525.0;
 T0 = 6.697374558 + (2400.051336*T) + (0.000025862*T*T);

 if (T0 > 24.0)
 {
  while( T0 > 24) T0 -= 24;
 }
 else if ( T0 < 0.0)
 {
  while( T0 < 0.0) T0 +=24;
 }

 UT *= 1.002737909;
 UT += T0;

 if (UT > 24.0)
 {
  while( UT > 24) UT -= 24;
 }
 else if ( UT < 0.0)
 {
  while( UT < 0.0) UT +=24;
 }
return(UT);
}

/* ====================================================== */
/* = Fonction    : Calcul POSITION PLANETES             = */
/* = Entree      : objet date                           = */
/* = Sortie      : structure position                   = */
/* ====================================================== */
POSITION CalculPosition(struct ROT_STRUCT *p_rot)
{
 POSITION p;

 struct DMS f2dms;
 struct DMS *p_f2dms = &f2dms;

 double T;  // temps sidral
 double t;  // temps sidral
 double o;  // obliguit de l'cliptique
 double L;  // Longitude moyenne
 double M;  // anomalie moyenne
 double E;  // anomalie excentrique
 double e;  // exentricit
 double a;  // demi grand axe
 double v;  // anomalie vrai
 double R;  // Rayon vecteur en UA

 double erreur = 1.0;

 double A;			// Ascention droite
 double D;			// Dclinaison
 double i,w,W;
 double r,u,l,b;		// l,b,r  coordonnes cliptique hliocentriques; longitude l, latitude b, rayon r
 double lambda,delta,beta;

 double Lg; // Longitude vrai du soleil

 double cu, cP, cQ, cV, cZ;
 double cH, cS, cN, cT;
 double cA, cB, cC, cD;
 
 p.JJ=JJ(p_rot->jj, p_rot->mm, p_rot->aa, p_rot->hrs, p_rot->min, p_rot->sec);

 double J=floor(p.JJ+0.5)-0.5;			// jour julien  0H00
 double j=p.JJ-J;				// portion du jour julien restant
 T=(J-2415020.0)/36525.0;
 t=0.276919398+100.0021359*T+0.000001075*T*T;	// temps sidral  0H00 GMT exprim en rvolutions 

 t=t+j*1.002737908;				// temps sidral  l'heure correspondant
						//  la portion de jour jj
 t=fmod(t,1);
 t=t*2*PI; 					// en radian
 p.t=t;

 T=(JJ(p_rot->jj, p_rot->mm, p_rot->aa, p_rot->hrs, p_rot->min, p_rot->sec)-2415020.0)/36525.0;
 o=23.452294-0.0130125*T-0.00000164*T*T+0.000000503*T*T*T; o=fmod(o,360.0); o=o*PI/180.0;
 L=279.69668+36000.76892*T+0.0003025*T*T; L=fmod(L,360.0); L=L*PI/180.0;
 M=358.47583+35999.04975*T-0.000150*T*T-0.0000033*T*T*T;  M=fmod(M,360); M=M*PI/180.0;
 e=0.01675104-0.0000418*T-0.000000126*T*T;
 
 E=M;

 while (erreur > 1e-12) { erreur = (M+e*sin(E)-E)/(1-e*cos(E)); E += erreur; }

 v=2*atan(sqrt((1+e)/(1-e))*tan(E/2));
 a=1.0000002;
 R=a*(1-e*cos(E));
 Lg=L+v-M;

  if (!p_rot->obj)
  {
   p.Asc  = atan2(cos(o)*sin(Lg),cos(Lg));
   p.Dec  = asin(sin(o)*sin(Lg));
   p.Dist = R;
  }
  else
  {
   switch (p_rot->obj) {

   case MERCURE:
   L= 178.179078 + 149474.07078 *T + 0.0003011 *T*T;
   a=  0.3870986;
   e=  0.20561421 + 0.00002046 *T  - 0.000000030 *T*T;
   i=  7.002881 + 0.0018608 *T  - 0.0000183 *T*T;
   w= 28.753753 + 0.3702806 *T  + 0.0001208 *T*T;
   W= 47.145944 + 1.1852083 *T  + 0.0001739 *T*T;
   break;

   case VENUS:
   L= 342.767053 + 58519.21191 *T + 0.0003097 *T*T;
   a=  0.7233316;
   e=  0.00682069 - 0.00004774 *T  + 0.000000091 *T*T;
   i=  3.393631 + 0.0010058 *T  - 0.0000010 *T*T;
   w= 54.384186 + 0.5081861 *T  - 0.0013864 *T*T;
   W= 75.779647 + 0.8998500 *T  + 0.0004100 *T*T;
   break;

   case MARS:
   L= 293.737334 + 19141.69551 *T + 0.0003107 *T*T;
   a=  1.5236883;
   e=  0.09331290 + 0.000092064 *T - 0.000000077 *T*T;
   i=  1.850333 - 0.0006750 *T  + 0.0000126 *T*T;
   w= 285.431761 + 1.0697667 *T  + 0.0001313 *T*T + 0.00000414 *T*T*T;
   W= 48.786442 + 0.7709917 *T  - 0.0000014 *T*T - 0.00000533 *T*T*T;
   break;

   case JUPITER:
   L= 238.049257 + 3036.301986 *T + 0.0003347 *T*T - 0.00000165 *T*T*T;
   a=  5.202561;
   e=  0.04833475 + 0.000164180 *T - 0.0000004676 *T*T - 0.0000000017 *T*T*T;
   i=  1.308736 - 0.0056961 *T  + 0.0000039 *T*T;
   w= 273.277558 + 0.5994317 *T  + 0.00070405 *T*T + 0.00000508 *T*T*T;
   W= 99.443414 + 1.0105300 *T  + 0.00035222 *T*T - 0.00000851 *T*T*T;
   break;

   case SATURNE:
   L= 266.564377 + 1223.509884 *T + 0.0003245 *T*T - 0.0000058 *T*T*T;
   a=  9.554747;
   e=  0.05589232 - 0.00034550 *T  - 0.000000728 *T*T + 0.00000000074 *T*T*T;
   i=  2.492519 - 0.0039189 *T  - 0.00001549 *T*T + 0.00000004 *T*T*T;
   w= 338.307800 + 1.0852207 *T  + 0.00097854 *T*T + 0.00000992 *T*T*T;
   W= 112.790414 + 0.8731951 *T  - 0.00015218 *T*T - 0.00000531 *T*T*T;
   break;
  }

  M=L-w-W;
 
  //----------- perturbation  par attractions mutuelles ---------------
  switch (p_rot->obj)
  {
   case JUPITER:
   cu = 0.1  + T/5;
   cP = 237.475 + 3034.9061 *T; // angle en degrs
   cQ = 265.916 + 1222.1139 *T; // angle en degrs
   cV = 5*cQ - 2*cP;		// angle en degrs
   cZ = cQ - cP;		// angle en degrs

    cP=fmod(cP,360);cP=deg2rad(cP);
    cQ=fmod(cQ,360);cQ=deg2rad(cQ);
    cV=fmod(cV,360);cV=deg2rad(cV);
    cZ=fmod(cZ,360);cZ=deg2rad(cZ);

  // perturbation dans la longitude moyenne, en degrs:
    cA = + (0.3314 - 0.0103*cu - 0.0047*cu*cu)*sin(cV) + (0.0032 - 0.0644*cu + 0.0021*cu*cu)*cos(cV) +  0.0136*sin(cZ) + 0.0185*sin(2*cZ) + 0.0067*sin(3*cZ) + (0.0073*sin(cZ) + 0.0064*sin(2*cZ) - 0.0338*cos(cZ))*sin(cQ) - (0.0357*sin(cZ) + 0.0063*cos(cZ) + 0.0067*cos(2*cZ))*cos(cQ);
  // perturbation dans l'excentricit, en units de la sixime dcimale:
    cB = + (361 + 13*cu)*sin(cV) + (129 - 58*cu)*cos(cV) + (128*cos(cZ) - 676*sin(cZ) - 111*sin(2*cZ))*sin(cQ) + (146*sin(cZ) - 82 + 607*cos(cZ) + 99*cos(2*cZ) + 51*cos(3*cZ))*cos(cQ) - (96*sin(cZ) + 100*cos(cZ))*sin(2*cQ) - (96*sin(cZ) - 102*cos(cZ))*cos(2*cQ);
  // perturbation dans la longitude du prihlie, en degrs:
    cC = + (0.0072 - 0.0031*cu)*sin(cV) - 0.0204*cos(cV) + (0.0073*sin(cZ) + 0.0340*cos(cZ) + 0.0056*cos(2*cZ))*sin(cQ) + (0.0378*sin(cZ) + 0.0062*sin(2*cZ) - 0.0066*cos(cZ))*cos(cQ) - 0.0054*sin(cZ)*sin(2*cQ) + 0.0055*cos(cZ)*cos(2*cQ);
  // perturbation dans le demie-grand axe, en units de la sixime dcimale:
    cD = - 263*cos(cV) + 205*cos(cZ) + 693*cos(2*cZ) + 312*cos(3*cZ) + 299*sin(cZ)*sin(cQ) + (204*sin(2*cZ)-337*cos(cZ))*cos(cQ);
    L += cA;
    M += cA - cC/e;
    e += cB/1000000;
    a += cD/1000000;
    break;

  case SATURNE:
   cu = 0.1  + T/5;
   cP = 237.475 + 3034.9061 *T; // angle en degrs
   cQ = 265.916 + 1222.1139 *T; // angle en degrs
   cV = 5*cQ - 2*cP;    	// angle en degrs
   cZ = cQ - cP;		// angle en degrs

    cP=fmod(cP,360);cP=deg2rad(cP);
    cQ=fmod(cQ,360);cQ=deg2rad(cQ);
    cV=fmod(cV,360);cV=deg2rad(cV);
    cZ=fmod(cZ,360);cZ=deg2rad(cZ);

  // perturbation dans la longitude moyenne, en degrs:
   cA = + (- 0.8142 + 0.0181*cu + 0.0167*cu*cu)*sin(cV) + (- 0.0105 + 0.1609*cu - 0.0041*cu*cu)*cos(cV)  -  0.1488*sin(cZ) - 0.0408*sin(2*cZ) - 0.0152*sin(3*cZ) + (0.0089*sin(cZ) - 0.0165*sin(2*cZ))*sin(cQ) + (0.0813*cos(cZ) + 0.0150*cos(2*cZ))*sin(cQ) + (0.0856*sin(cZ) + 0.0253*cos(cZ) + 0.0144*cos(2*cZ))*cos(cQ) + 0.0092*sin(2*cZ)*sin(2*cQ);
  // perturbation dans l'excentricit, en units de la sixime dcimale:
   cB = + (- 793 + 255*cu)*sin(cV) + (1338 + 123*cu)*cos(cV) + 1241*sin(cQ) + (39-62*cu)*sin(cZ)*sin(cQ) + (2660*cos(cZ) - 469*cos(2*cZ) - 187*cos(3*cZ) - 82*cos(4*cZ))*sin(cQ) - (1270*sin(cZ) + 420*sin(2*cZ) +150*sin(3*cZ))*cos(cQ) - 62*sin(4*cZ)*cos(cQ) + (221*sin(cZ) - 221*sin(2*cZ) - 57*sin(3*cZ))*sin(2*cQ) - (278*cos(cZ) - 202*cos(2*cZ))*sin(2*cQ) - (284*sin(cZ) + 159*cos(cZ))*cos(2*cQ) + (216*cos(2*cZ) + 56*cos(3*cZ))*cos(2*cQ);
  // perturbation dans la longitude du prihlie, en degrs:
   cC = + (0.0771 + 0.0072*cu)*sin(cV) + (0.0458 - 0.0148*cu)*cos(cV) - (0.0758*sin(cZ) + 0.0248*sin(2*cZ) + 0.0086*sin(3*cZ))*sin(cQ) - (0.0726 + 0.1504*cos(cZ) - 0.0269*cos(2*cZ) - 0.0101*cos(3*cZ))*cos(cQ) - (0.0136*sin(cZ) - 0.0136*cos(2*cZ))*sin(2*cQ) - (0.0137*sin(cZ) - 0.0120*sin(2*cZ))*cos(2*cQ) + (0.0149*cos(cZ) - 0.0131*cos(2*cZ))*cos(2*cQ);
  // perturbation dans le demie-grand axe, en units de la sixime dcimale:
   cD = + 2933*cos(cV) + 33629*cos(cZ) - 3081*cos(2*cZ) - 1423*cos(3*cZ) - 671*cos(4*cZ) + (1098 - 2812*sin(cZ) + 688*sin(2*cZ))*sin(cQ) + (2138*cos(cZ) - 999*cos(2*cZ) - 642*cos(3*cZ))*sin(cQ) - 890*cos(cQ) + (2206*sin(cZ) - 1590*sin(2*cZ) - 647*sin(3*cZ))*cos(cQ) + (2885*cos(cZ) + 2172*cos(2*cZ))*cos(cQ) - 778*cos(cZ)*sin(2*cQ) - 856*sin(cZ)*cos(2*cQ);
   L += cA;
   M += cA - cC/e;
   e += cB/1000000.0;
   a += cD/1000000.0;
   break;
  }
 
  L=fmod(L,360);L=deg2rad(L);
  i=fmod(i,360);i=deg2rad(i);
  w=fmod(w,360);w=deg2rad(w);
  W=fmod(W,360);W=deg2rad(W);
  M=fmod(M,360);M=deg2rad(M);

  E=M;
  erreur=1.0;

  while (erreur > 1e-12) { erreur = (M+e*sin(E)-E)/(1-e*cos(E)); E += erreur; }

  v=2*atan(sqrt((1+e)/(1-e))*tan(E/2));
  r=a*(1-e*cos(E));
  u=L+v-M-W;
  l = W + atan2(cos(i)*sin(u),cos(u));
  b=asin(sin(u)*sin(i));

  //Lg et R et o
  lambda=Lg+atan2(r*cos(b)*sin(l-Lg), r*cos(b)*cos(l-Lg) + R);
  delta=sqrt(R*R+r*r+2*r*R*cos(b)*cos(l-Lg));
  beta=asin(r/delta*sin(b));
   A = atan2(sin(lambda)*cos(o)-tan(beta)*sin(o), cos(lambda));
   D = asin(sin(beta)*cos(o)+cos(beta)*sin(o)*sin(lambda));
   p.Asc  = A;
   p.Dec  = D;
   p.Dist = delta;
 }

/* *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** */
/* debut des calculs de la position apparente de l'objet vise  *** *** *** *** *** *** */
/* *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** */

 p.Asc = rad2deg(p.Asc / 15.0);
 p.Dec = rad2deg(p.Dec);

 // calcul temps sideral pour greenwich avec le temps UT au temps corrige
 double sideral = STG_H(p_rot);
 
 // calcul du temps sideral local pour le lieu d'observation (-)EST Greenwich ou (+)OUEST
 sideral -= p_rot->lon / 15.0;
 if(sideral > 24.0) sideral -= 24.0;
 if(sideral < 0.0 ) sideral += 24.0;

 // angle horaire = temps local sideral - ascension droite objet
 double H = (sideral - p.Asc);
 if (H < 0.0) H += 24.0;

 p.Asc = deg2rad(p.Asc);
 p.Dec = deg2rad(p.Dec);
 p_rot->lat = deg2rad(p_rot->lat);
 p_rot->lon = deg2rad(p_rot->lon);

 // degrees conversion from HMS
 H *= 15.0; 
 H = deg2rad(H);

 if( p_rot->ht == -1)  // -a non renseigne azimuth egal nul
 {
  p_rot->ht = ( sin (p.Dec)*sin(p_rot->lat) + cos(p.Dec)*cos(p_rot->lat)*cos(H) );
  p_rot->ht = asin(p_rot->ht);
 }

 if( p_rot->az == -1)  // -h non renseigne hauteur egal nul
 {
  p_rot->az = ( sin(p.Dec)-(sin(p_rot->lat)*sin(p_rot->ht)) ) / ( cos(p_rot->lat)*cos(p_rot->ht) );
  //  while(p_rot->az < 0.0 )   p_rot->az += (2*PI);
  //  while(p_rot->az > (2*PI)) p_rot->az -= (2*PI);
  p_rot->az = acos(p_rot->az);
  if( sin(H) > 0.0 ) p_rot->az = (2*PI) - p_rot->az;
 }

 // conversion en degres !
 p.Asc = rad2deg(p.Asc);
 // normalisation RA
 if( p.Asc < 0.0 ) { p.Asc = p.Asc + 24.0; }
 p.Dec = rad2deg(p.Dec);
 p_rot->asc = p.Asc;
 p_rot->dec = p.Dec;
 p_rot->ht = rad2deg(p_rot->ht);
 p_rot->az = rad2deg(p_rot->az);
 p_rot->lat = rad2deg(p_rot->lat);
 p_rot->lon = rad2deg(p_rot->lon);

/* *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** */
/* fin des calculs de positions ** *** *** *** *** *** *** *** *** *** *** *** *** *** */
/* *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** */

 return p;
} 


/* ====================================================== */
/* = Fonction    : Calcul Angle par image               = */
/* = Entree      : structure                            = */
/* = Sortie      : double                               = */
/* ====================================================== */
double RImage(struct ROT_STRUCT *p_rotA)
{
 POSITION p;

 struct DMS f2dms;
 struct DMS *p_f2dms = &f2dms;

 double Ri  = 0.0;
 double Rch = 0.0;
 double B   = 0.0;

 // calcul du temps de pose theorique par image
 if( p_rotA->sec == -1.0 ) { p_rotA->sec = 0.0; }
 else
 {
  B = (p_rotA->nimg/p_rotA->img)/p_rotA->nimg;

  // Conversion pour incrementation a TU + Pose en secondes
  Float2DMS((B/3600.0), p_f2dms);
  p_rotA->hrs = p_rotA->hrs + p_f2dms->d;
  p_rotA->min = p_rotA->min + p_f2dms->m;
  p_rotA->sec = p_rotA->sec + p_f2dms->s;
 }

 // test de date sur incrementation
 if( p_rotA->sec > 59.0 )
 {
  p_rotA->sec = 0.0;
  p_rotA->min = p_rotA->min + 1;
 }

 if( p_rotA->min > 59 )
 {
  p_rotA->min = 0;
  p_rotA->hrs = p_rotA->hrs + 1;
 }
  
 if( p_rotA->hrs > 23 )
 {
  p_rotA->hrs = 0;
  p_rotA->jj = p_rotA->jj + 1.0;
 }

 // verifier le nombre de jours du mois
 if( Int_Float(p_rotA->jj) > (double)NbDaysFMonth(p_rotA->mm, p_rotA->aa) )
 {
  p_rotA->jj = 1;
  p_rotA->mm += 1;
  // verifie l'annee en cours
  if( p_rotA->mm > 12 ) 
  {
   p_rotA->mm = 1;
   p_rotA->aa += 1;
  }
 }

 p_rotA->jj = Day2Float(Int_Float(p_rotA->jj), p_rotA->hrs, p_rotA->min, p_rotA->sec);

 // optimisation de la position pour estimation rotation
 if( p_rotA->obj != -1 )
 {
  p_rotA->az = -1; //__ deg2rad(p_rotA->az);
  p_rotA->ht = -1; //__ deg2rad(p_rotA->ht);
  p = CalculPosition(p_rotA);
 }

 // calcul de la rotation de champ
 Rch = rot_moy(p_rotA->lat, p_rotA->az, p_rotA->ht);

 if( p_rotA->trig )			// Rotation dans le sens trigo (+)
  Ri = fabs( (fabs(Rch) / 3600.0) / p_rotA->img ) ;
 else					// Rotation dans le sens horaire (-)
  Ri = ( (Rch / 3600.0) / p_rotA->img);

 return (Ri);
}

